@@ -19,10 +19,11 @@
  * 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
+#include <glib/gi18n-lib.h>
+
 #include <stdlib.h>
 #include <string.h>
-#include <glib/gi18n-lib.h>
 #include <dbus/dbus-glib.h>
 
 #include "gdu-private.h"
@@ -53,6 +54,7 @@ struct _GduDrivePrivate
 {
         GduDevice *device;
         GduPool *pool;
+        GduPresentable *enclosing_presentable;
         gchar *id;
 };
 
@@ -66,10 +68,22 @@ G_DEFINE_TYPE_WITH_CODE (GduDrive, gdu_drive, G_TYPE_OBJECT,
 static void device_job_changed (GduDevice *device, gpointer user_data);
 static void device_changed (GduDevice *device, gpointer user_data);
 
-static gboolean gdu_drive_has_unallocated_space_real (GduDrive        *drive,
-                                                      gboolean        *out_whole_disk_is_unitialized,
-                                                      guint64         *out_largest_segment,
-                                                      GduPresentable **out_presentable);
+static gboolean gdu_drive_can_create_volume_real (GduDrive        *drive,
+                                                  gboolean        *out_is_uninitialized,
+                                                  guint64         *out_largest_contiguous_free_segment,
+                                                  guint64         *out_total_free,
+                                                  GduPresentable **out_presentable);
+
+static void gdu_drive_create_volume_real (GduDrive              *drive,
+                                          guint64                size,
+                                          const gchar           *name,
+                                          GduCreateVolumeFlags   flags,
+                                          GAsyncReadyCallback    callback,
+                                          gpointer               user_data);
+
+static GduVolume *gdu_drive_create_volume_finish_real (GduDrive              *drive,
+                                                       GAsyncResult          *res,
+                                                       GError               **error);
 
 static void
 gdu_drive_finalize (GduDrive *drive)
@@ -85,6 +99,9 @@ gdu_drive_finalize (GduDrive *drive)
         if (drive->priv->pool != NULL)
                 g_object_unref (drive->priv->pool);
 
+        if (drive->priv->enclosing_presentable != NULL)
+                g_object_unref (drive->priv->enclosing_presentable);
+
         g_free (drive->priv->id);
 
         if (G_OBJECT_CLASS (parent_class)->finalize)
@@ -100,7 +117,9 @@ gdu_drive_class_init (GduDriveClass *klass)
 
         obj_class->finalize = (GObjectFinalizeFunc) gdu_drive_finalize;
 
-        klass->has_unallocated_space = gdu_drive_has_unallocated_space_real;
+        klass->can_create_volume = gdu_drive_can_create_volume_real;
+        klass->create_volume = gdu_drive_create_volume_real;
+        klass->create_volume_finish = gdu_drive_create_volume_finish_real;
 
         g_type_class_add_private (klass, sizeof (GduDrivePrivate));
 }
@@ -194,25 +213,25 @@ gdu_drive_deactivate (GduDrive                *drive,
 }
 
 /**
- * gdu_drive_has_unallocated_space:
+ * gdu_drive_can_create_volume:
  * @drive: A #GduDrive.
- * @out_whole_disk_is_unitialized: Return location for whether @drive is uninitialized or %NULL.
- * @out_largest_segment: Return location biggest contigious free block of @drive or %NULL.
+ * @out_is_uninitialized: Return location for whether @drive is uninitialized or %NULL.
+ * @out_largest_segment: Return location for biggest contigious free block of @drive or %NULL.
+ * @out_total_free: Return location for total amount of free space on @drive or %NULL.
  * @out_presentable: Return location for the presentable that represents free space or %NULL. Free
  * with g_object_unref().
  *
- * This method computes the largest contigious free block of
- * unallocated space on @drive.
+ * This method checks if a new volume can be created on @drive.
  *
  * If @drive uses removable media and there is no media inserted,
  * %FALSE is returned.
  *
  * If @drive appears to be completely uninitialized (such as a hard
- * disk full of zeros), @out_whole_disk_is_unitialized is set to
- * %TRUE, the size of the media/disk is returned in
- * @out_largest_segment and %TRUE is returned. Note that this can
- * also happen if @drive contains signatures unknown to the operating system
- * so be careful.
+ * disk full of zeros), @out_is_unitialized is set to %TRUE, the size
+ * of the media/disk is returned in @out_largest_segment and %TRUE is
+ * returned. Note that this can also happen if @drive contains
+ * signatures unknown to the operating system so be careful and always
+ * prompt the user.
  *
  * If the disk is partitioned and unallocated space exists but no more
  * partitions can be created (due to e.g. four primary partitions on a
@@ -223,31 +242,113 @@ gdu_drive_deactivate (GduDrive                *drive,
  * #GduVolumeHole (if the disk is partitioned and has free space) or
  * the #GduDrive (if the disk is uninitialized).
  *
+ * You can use gdu_drive_create_volume() to create a volume.
+ *
  * Returns: %TRUE if @drive has unallocated space, %FALSE otherwise.
  */
 gboolean
-gdu_drive_has_unallocated_space (GduDrive        *drive,
-                                 gboolean        *out_whole_disk_is_unitialized,
-                                 guint64         *out_largest_segment,
-                                 GduPresentable **out_presentable)
+gdu_drive_can_create_volume (GduDrive        *drive,
+                             gboolean        *out_is_uninitialized,
+                             guint64         *out_largest_contiguous_free_segment,
+                             guint64         *out_total_free,
+                             GduPresentable **out_presentable)
 {
         GduDriveClass *klass = GDU_DRIVE_GET_CLASS (drive);
-
-        return klass->has_unallocated_space (drive,
-                                             out_whole_disk_is_unitialized,
-                                             out_largest_segment,
-                                             out_presentable);
+        return klass->can_create_volume (drive,
+                                         out_is_uninitialized,
+                                         out_largest_contiguous_free_segment,
+                                         out_total_free,
+                                         out_presentable);
 }
 
+/**
+ * gdu_drive_count_mbr_partitions:
+ * @drive: A #GduDrive.
+ * @out_num_primary_partitions: Return location for number of primary partitions.
+ * @out_has_extended_partition: Return location for number of extended partitions.
+ *
+ * Counts the number of primary partitions and figures out if there's an extended partition.
+ *
+ * Returns: %TRUE if @out_num_logical_partitions and @out_has_extended_partition is set, %FALSE otherwise.
+ */
+gboolean
+gdu_drive_count_mbr_partitions (GduDrive  *drive,
+                                guint     *out_num_primary_partitions,
+                                gboolean  *out_has_extended_partition)
+{
+        guint num_primary_partitions;
+        gboolean has_extended_partition;
+        gboolean ret;
+        GduDevice *device;
+        GduPool *pool;
+        GList *enclosed_presentables;
+        GList *l;
+
+        ret = FALSE;
+        num_primary_partitions = 0;
+        has_extended_partition = FALSE;
+        pool = NULL;
+        device = NULL;
+
+        device = gdu_presentable_get_device (GDU_PRESENTABLE (drive));
+        if (device == NULL)
+                goto out;
+
+        pool = gdu_presentable_get_pool (GDU_PRESENTABLE (drive));
+
+        if (!gdu_device_is_partition_table (device) ||
+            g_strcmp0 (gdu_device_partition_table_get_scheme (device), "mbr") != 0) {
+                goto out;
+        }
+
+        enclosed_presentables = gdu_pool_get_enclosed_presentables (pool, GDU_PRESENTABLE (drive));
+        for (l = enclosed_presentables; l != NULL; l = l->next) {
+                GduPresentable *ep = GDU_PRESENTABLE (l->data);
+
+                if (GDU_IS_VOLUME (ep)) {
+                        gint type;
+                        GduDevice *ep_device;
+
+                        ep_device = gdu_presentable_get_device (ep);
+
+                        type = strtol (gdu_device_partition_get_type (ep_device), NULL, 0);
+                        if (type == 0x05 || type == 0x0f || type == 0x85) {
+                                has_extended_partition = TRUE;
+                        }
+                        num_primary_partitions++;
+                        g_object_unref (ep_device);
+                }
+        }
+        g_list_foreach (enclosed_presentables, (GFunc) g_object_unref, NULL);
+        g_list_free (enclosed_presentables);
+
+        ret = TRUE;
+
+ out:
+        if (device != NULL)
+                g_object_unref (device);
+        if (pool != NULL)
+                g_object_unref (pool);
+
+        if (out_num_primary_partitions != NULL)
+                *out_num_primary_partitions = num_primary_partitions;
+        if (out_has_extended_partition != NULL)
+                *out_has_extended_partition = has_extended_partition;
+        return ret;
+}
+
+
 static gboolean
-gdu_drive_has_unallocated_space_real (GduDrive        *drive,
-                                      gboolean        *out_whole_disk_is_unitialized,
-                                      guint64         *out_largest_segment,
-                                      GduPresentable **out_presentable)
+gdu_drive_can_create_volume_real (GduDrive        *drive,
+                                  gboolean        *out_is_unitialized,
+                                  guint64         *out_largest_contiguous_free_segment,
+                                  guint64         *out_total_free,
+                                  GduPresentable **out_presentable)
 {
         GduDevice *device;
         GduPool *pool;
-        guint64 largest_segment;
+        guint64 largest_contiguous_free_segment;
+        guint64 total_free;
         gboolean whole_disk_uninitialized;
         GList *enclosed_presentables;
         GList *l;
@@ -256,7 +357,8 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
         guint64 size;
         GduPresentable *pres;
 
-        largest_segment = 0;
+        largest_contiguous_free_segment = 0;
+        total_free = 0;
         whole_disk_uninitialized = FALSE;
         ret = FALSE;
         device = NULL;
@@ -282,7 +384,8 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
          */
         if (!gdu_device_is_partition_table (device) && strlen (gdu_device_id_get_usage (device)) == 0) {
                 whole_disk_uninitialized = TRUE;
-                largest_segment = gdu_device_get_size (device);
+                largest_contiguous_free_segment = gdu_device_get_size (device);
+                total_free = gdu_device_get_size (device);
                 ret = TRUE;
                 pres = GDU_PRESENTABLE (drive);
                 goto out;
@@ -299,10 +402,13 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
                 if (GDU_IS_VOLUME_HOLE (ep)) {
                         size = gdu_presentable_get_size (ep);
 
-                        if (size > largest_segment) {
-                                largest_segment = size;
+                        if (size > largest_contiguous_free_segment) {
+                                largest_contiguous_free_segment = size;
                                 pres = ep;
                         }
+
+                        total_free += size;
+
                 } else if (GDU_IS_VOLUME (ep)) {
                         gint type;
                         GduDevice *ep_device;
@@ -323,10 +429,12 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
 
                                         if (GDU_IS_VOLUME_HOLE (lep)) {
                                                 size = gdu_presentable_get_size (lep);
-                                                if (size > largest_segment) {
-                                                        largest_segment = size;
+                                                if (size > largest_contiguous_free_segment) {
+                                                        largest_contiguous_free_segment = size;
                                                         pres = lep;
                                                 }
+
+                                                total_free += size;
                                         }
                                 }
                                 g_list_foreach (logical_partitions, (GFunc) g_object_unref, NULL);
@@ -338,10 +446,10 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
         g_list_foreach (enclosed_presentables, (GFunc) g_object_unref, NULL);
         g_list_free (enclosed_presentables);
 
-        ret = (largest_segment > 0);
+        ret = (largest_contiguous_free_segment > 0);
 
         /* Now igure out if the partition table is full (e.g. four primary partitions already) and
-         * return %FALSE and non-zero @out_largest_segment
+         * return %FALSE and non-zero @out_largest_contiguous_free_segment
          */
         if (g_strcmp0 (gdu_device_partition_table_get_scheme (device), "mbr") == 0 &&
             gdu_device_partition_table_get_count (device) == 4 &&
@@ -355,11 +463,14 @@ gdu_drive_has_unallocated_space_real (GduDrive        *drive,
         if (pool != NULL)
                 g_object_unref (pool);
 
-        if (out_largest_segment != NULL)
-                *out_largest_segment = largest_segment;
+        if (out_largest_contiguous_free_segment != NULL)
+                *out_largest_contiguous_free_segment = largest_contiguous_free_segment;
 
-        if (out_whole_disk_is_unitialized != NULL)
-                *out_whole_disk_is_unitialized = whole_disk_uninitialized;
+        if (out_total_free != NULL)
+                *out_total_free = total_free;
+
+        if (out_is_unitialized != NULL)
+                *out_is_unitialized = whole_disk_uninitialized;
 
         if (out_presentable != NULL) {
                 *out_presentable = (pres != NULL ? g_object_ref (pres) : NULL);
@@ -391,14 +502,18 @@ device_job_changed (GduDevice *device, gpointer user_data)
 }
 
 GduDrive *
-_gdu_drive_new_from_device (GduPool *pool, GduDevice *device)
+_gdu_drive_new_from_device (GduPool *pool, GduDevice *device, GduPresentable *enclosing_presentable)
 {
         GduDrive *drive;
 
         drive = GDU_DRIVE (g_object_new (GDU_TYPE_DRIVE, NULL));
         drive->priv->device = g_object_ref (device);
         drive->priv->pool = g_object_ref (pool);
-        drive->priv->id = g_strdup_printf ("drive_%s", gdu_device_get_device_file (drive->priv->device));
+        drive->priv->enclosing_presentable =
+                enclosing_presentable != NULL ? g_object_ref (enclosing_presentable) : NULL;
+        drive->priv->id = g_strdup_printf ("drive_%s_enclosed_by_%s",
+                                           gdu_device_get_device_file (drive->priv->device),
+                                           enclosing_presentable != NULL ? gdu_presentable_get_id (enclosing_presentable) : "(none)");
 
         g_signal_connect (device, "changed", (GCallback) device_changed, drive);
         g_signal_connect (device, "job-changed", (GCallback) device_job_changed, drive);
@@ -423,6 +538,9 @@ gdu_drive_get_device (GduPresentable *presentable)
 static GduPresentable *
 gdu_drive_get_enclosing_presentable (GduPresentable *presentable)
 {
+        GduDrive *drive = GDU_DRIVE (presentable);
+        if (drive->priv->enclosing_presentable != NULL)
+                return g_object_ref (drive->priv->enclosing_presentable);
         return NULL;
 }
 
@@ -559,7 +677,7 @@ gdu_drive_get_name (GduPresentable *presentable)
         is_rotational = gdu_device_drive_get_is_rotational (drive->priv->device);
 
         if (has_media && size > 0) {
-                strsize = gdu_util_get_size_for_display (size, FALSE);
+                strsize = gdu_util_get_size_for_display (size, FALSE, FALSE);
         }
 
         if (is_removable) {
@@ -619,11 +737,10 @@ gdu_drive_get_name (GduPresentable *presentable)
                 } else {
                         if (is_rotational) {
                                 if (strsize != NULL) {
-                                        /* Translators: This string is used to describe a hard disk. The first %s is
-                                         * the size of the drive e.g. '45 GB'.
+                                        /* Translators: This string is used to describe a hard disk.
+                                         * The first %s is the size of the drive e.g. '45 GB'.
                                          */
-                                        g_string_append_printf (result, _("%s Hard Disk"),
-                                                                strsize);
+                                        g_string_append_printf (result, _("%s Hard Disk"), strsize);
                                 } else {
                                         /* Translators: This string is used to describe a hard disk where the size
                                          * is not known.
@@ -678,7 +795,7 @@ gdu_drive_get_description (GduPresentable *presentable)
         if (is_removable) {
                 if (has_media && size > 0) {
                         gchar *strsize;
-                        strsize = gdu_util_get_size_for_display (size, FALSE);
+                        strsize = gdu_util_get_size_for_display (size, FALSE, FALSE);
                         /* Translators: This string is the description of a drive. The first %s is the
                          * size of the inserted media, for example '45 GB'.
                          */
@@ -800,6 +917,7 @@ gdu_drive_get_icon (GduPresentable *presentable)
         const char *presentation_icon_name;
         gchar **drive_media_compat;
         gboolean is_removable;
+        GIcon *icon;
 
         connection_interface = gdu_device_drive_get_connection_interface (drive->priv->device);
         is_removable = gdu_device_is_removable (drive->priv->device);
@@ -879,7 +997,39 @@ gdu_drive_get_icon (GduPresentable *presentable)
                         name = "drive-harddisk";
         }
 
-        return g_themed_icon_new_with_default_fallbacks (name);
+        /* Attach a MP emblem if it's a multipathed device or a path for a multipathed device */
+        icon = g_themed_icon_new_with_default_fallbacks (name);
+        if (gdu_device_is_linux_dmmp (drive->priv->device)) {
+                GEmblem *emblem;
+                GIcon *padlock;
+                GIcon *emblemed_icon;
+
+                padlock = g_themed_icon_new ("gdu-emblem-mp");
+                emblem = g_emblem_new_with_origin (padlock, G_EMBLEM_ORIGIN_DEVICE);
+
+                emblemed_icon = g_emblemed_icon_new (icon, emblem);
+                g_object_unref (icon);
+                icon = emblemed_icon;
+
+                g_object_unref (padlock);
+                g_object_unref (emblem);
+        } else if (gdu_device_is_linux_dmmp_component (drive->priv->device)) {
+                GEmblem *emblem;
+                GIcon *padlock;
+                GIcon *emblemed_icon;
+
+                padlock = g_themed_icon_new ("gdu-emblem-mp-component");
+                emblem = g_emblem_new_with_origin (padlock, G_EMBLEM_ORIGIN_DEVICE);
+
+                emblemed_icon = g_emblemed_icon_new (icon, emblem);
+                g_object_unref (icon);
+                icon = emblemed_icon;
+
+                g_object_unref (padlock);
+                g_object_unref (emblem);
+        }
+
+        return icon;
 }
 
 static guint64
@@ -932,3 +1082,387 @@ gdu_drive_presentable_iface_init (GduPresentableIface *iface)
         iface->is_recognized             = gdu_drive_is_recognized;
 }
 
+void
+_gdu_drive_rewrite_enclosing_presentable (GduDrive *drive)
+{
+        if (drive->priv->enclosing_presentable != NULL) {
+                const gchar *enclosing_presentable_id;
+                GduPresentable *new_enclosing_presentable;
+
+                enclosing_presentable_id = gdu_presentable_get_id (drive->priv->enclosing_presentable);
+
+                new_enclosing_presentable = gdu_pool_get_presentable_by_id (drive->priv->pool,
+                                                                            enclosing_presentable_id);
+                if (new_enclosing_presentable == NULL) {
+                        g_warning ("Error rewriting enclosing_presentable for %s, no such id %s",
+                                   drive->priv->id,
+                                   enclosing_presentable_id);
+                        goto out;
+                }
+
+                g_object_unref (drive->priv->enclosing_presentable);
+                drive->priv->enclosing_presentable = new_enclosing_presentable;
+        }
+
+ out:
+        ;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static void
+create_volume_partition_create_cb (GduDevice  *device,
+                                   gchar      *created_device_object_path,
+                                   GError     *error,
+                                   gpointer    user_data)
+{
+        GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (user_data);
+
+        if (error != NULL) {
+                g_simple_async_result_set_from_error (simple, error);
+                g_error_free (error);
+        } else {
+                GduDevice *d;
+                GduPool *pool;
+                GduPresentable *volume;
+
+                pool = gdu_device_get_pool (device);
+                d = gdu_pool_get_by_object_path (pool, created_device_object_path);
+                g_assert (d != NULL);
+
+                volume = gdu_pool_get_volume_by_device (pool, d);
+                g_assert (volume != NULL);
+
+                g_simple_async_result_set_op_res_gpointer (simple, volume, g_object_unref);
+
+                g_object_unref (pool);
+                g_object_unref (d);
+        }
+        g_simple_async_result_complete_in_idle (simple);
+}
+
+static void
+gdu_drive_create_volume_real_internal (GduDrive              *drive,
+                                       guint64                size,
+                                       const gchar           *name,
+                                       GduCreateVolumeFlags   flags,
+                                       GSimpleAsyncResult    *simple);
+
+static void
+create_volume_partition_table_create_cb (GduDevice  *device,
+                                         GError     *error,
+                                         gpointer    user_data)
+{
+        GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (user_data);
+
+        if (error != NULL) {
+                g_simple_async_result_set_from_error (simple, error);
+                g_simple_async_result_complete_in_idle (simple);
+                g_error_free (error);
+        } else {
+                GduDrive              *drive;
+                guint64                size;
+                const gchar           *name;
+                GduCreateVolumeFlags   flags;
+
+                drive = GDU_DRIVE (g_async_result_get_source_object (G_ASYNC_RESULT (simple)));
+                size = (* ((guint64 *) g_object_get_data (G_OBJECT (simple), "gdu-size")));
+                name = g_object_get_data (G_OBJECT (simple), "gdu-name");
+                flags = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (simple), "gdu-flags"));
+
+                /* now that we have a partition table... try creating the volume again */
+                gdu_drive_create_volume_real_internal (drive,
+                                                       size,
+                                                       name,
+                                                       flags,
+                                                       simple);
+
+                g_object_unref (drive);
+        }
+}
+
+static void
+gdu_drive_create_volume_real (GduDrive              *drive,
+                              guint64                size,
+                              const gchar           *name,
+                              GduCreateVolumeFlags   flags,
+                              GAsyncReadyCallback    callback,
+                              gpointer               user_data)
+{
+        GSimpleAsyncResult *simple;
+
+        g_return_if_fail (GDU_IS_DRIVE (drive));
+
+        simple = g_simple_async_result_new (G_OBJECT (drive),
+                                            callback,
+                                            user_data,
+                                            gdu_drive_create_volume);
+
+        g_object_set_data_full (G_OBJECT (simple), "gdu-size", g_memdup (&size, sizeof (guint64)), g_free);
+        g_object_set_data_full (G_OBJECT (simple), "gdu-name", g_strdup (name), g_free);
+        g_object_set_data (G_OBJECT (simple), "gdu-flags", GINT_TO_POINTER (flags));
+
+        gdu_drive_create_volume_real_internal (drive,
+                                               size,
+                                               name,
+                                               flags,
+                                               simple);
+}
+
+static void
+gdu_drive_create_volume_real_internal (GduDrive              *drive,
+                                       guint64                size,
+                                       const gchar           *name,
+                                       GduCreateVolumeFlags   flags,
+                                       GSimpleAsyncResult    *simple)
+{
+        GduPresentable *p;
+        GduDevice *d;
+        gboolean whole_disk_is_uninitialized;
+        guint64 largest_segment;
+
+        if (!gdu_drive_can_create_volume (drive,
+                                          &whole_disk_is_uninitialized,
+                                          &largest_segment,
+                                          NULL, /* total_free */
+                                          &p)) {
+                g_simple_async_result_set_error (simple,
+                                                 GDU_ERROR,
+                                                 GDU_ERROR_FAILED,
+                                                 "Insufficient space");
+                g_simple_async_result_complete_in_idle (simple);
+                g_object_unref (simple);
+                goto out;
+        }
+
+        g_assert (p != NULL);
+
+        d = gdu_presentable_get_device (GDU_PRESENTABLE (drive));
+
+        if (GDU_IS_VOLUME_HOLE (p)) {
+                guint64 offset;
+                const gchar *scheme;
+                const gchar *type;
+                gchar *label;
+
+                offset = gdu_presentable_get_offset (p);
+
+                scheme = gdu_device_partition_table_get_scheme (d);
+                type = "";
+                label = NULL;
+                if (g_strcmp0 (scheme, "mbr") == 0) {
+                        if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_MD) {
+                                type = "0xfd";
+                        } else if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_LVM2) {
+                                type = "0x8e";
+                        }
+                } else if (g_strcmp0 (scheme, "gpt") == 0) {
+                        if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_MD) {
+                                type = "A19D880F-05FC-4D3B-A006-743F0F84911E";
+                                /* Limited to 36 UTF-16LE characters according to on-disk format..
+                                 * Since a RAID array name is limited to 32 chars this should fit */
+                                if (name != NULL)
+                                        label = g_strdup_printf ("RAID: %s", name);
+                                else
+                                        label = g_strdup_printf ("RAID component");
+                        } else if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_LVM2) {
+                                type = "E6D6D379-F507-44C2-A23C-238F2A3DF928";
+                                /* Limited to 36 UTF-16LE characters according to on-disk format..
+                                 * TODO: ensure name is shorter than or equal to 32 characters */
+                                if (name != NULL)
+                                        label = g_strdup_printf ("LVM2: %s", name);
+                                else
+                                        label = g_strdup_printf ("LVM2 component");
+                        }
+                } else if (g_strcmp0 (scheme, "apt") == 0) {
+                        type = "Apple_Unix_SVR2";
+                        if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_MD) {
+                                if (name != NULL)
+                                        label = g_strdup_printf ("RAID: %s", name);
+                                else
+                                        label = g_strdup_printf ("RAID component");
+                        } else if (flags & GDU_CREATE_VOLUME_FLAGS_LINUX_LVM2) {
+                                if (name != NULL)
+                                        label = g_strdup_printf ("LVM2: %s", name);
+                                else
+                                        label = g_strdup_printf ("LVM2 component");
+                        }
+                }
+
+                gdu_device_op_partition_create (d,
+                                                offset,
+                                                size,
+                                                type,
+                                                label != NULL ? label : "",
+                                                NULL,
+                                                "",
+                                                "",
+                                                "",
+                                                FALSE,
+                                                create_volume_partition_create_cb,
+                                                simple);
+                g_free (label);
+
+        } else {
+
+                /* otherwise the whole disk must be uninitialized... */
+                g_assert (whole_disk_is_uninitialized);
+
+                /* so create a partition table...
+                 * (TODO: take a flag to determine what kind of partition table to create)
+                 */
+                gdu_device_op_partition_table_create (d,
+                                                      "mbr",
+                                                      create_volume_partition_table_create_cb,
+                                                      simple);
+        }
+
+ out:
+        if (d != NULL)
+                g_object_unref (d);
+        if (p != NULL)
+                g_object_unref (p);
+}
+
+static GduVolume *
+gdu_drive_create_volume_finish_real (GduDrive              *drive,
+                                     GAsyncResult          *res,
+                                     GError               **error)
+{
+        GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (res);
+        GduVolume *ret;
+
+        g_return_val_if_fail (GDU_IS_DRIVE (drive), NULL);
+        g_return_val_if_fail (res != NULL, NULL);
+
+        g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == gdu_drive_create_volume);
+
+        ret = NULL;
+        if (g_simple_async_result_propagate_error (simple, error))
+                goto out;
+
+        ret = GDU_VOLUME (g_simple_async_result_get_op_res_gpointer (simple));
+
+ out:
+        return ret;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+/**
+ * gdu_drive_create_volume:
+ * @drive: A #GduDrive.
+ * @size: The size of the volume to create.
+ * @name: A name for the volume.
+ * @flags: Flags describing what kind of volume to create
+ * @callback: Function to call when the result is ready.
+ * @user_data: User data to pass to @callback.
+ *
+ * High-level method for creating a new volume on @drive of size @size
+ * using @name and @flags as influential hints.
+ *
+ * Depending on the actual type of @drive, different things may happen
+ * - if @drive represents a partitioned drive, then a new partition
+ * will be created (and if the partitioning scheme supports partition
+ * labels @name will be used as the label). If @drive is completely
+ * uninitialized, it may (or may not) be partitioned.
+ *
+ * If @drive represents a LVM2 volume group, a logical volume may be
+ * created (with @name being used as LV name).
+ *
+ * This is an asynchronous operation. When the result of the operation
+ * is ready, @callback will be invoked.
+ */
+void
+gdu_drive_create_volume (GduDrive              *drive,
+                         guint64                size,
+                         const gchar           *name,
+                         GduCreateVolumeFlags   flags,
+                         GAsyncReadyCallback    callback,
+                         gpointer               user_data)
+{
+        GduDriveClass *klass = GDU_DRIVE_GET_CLASS (drive);
+        klass->create_volume (drive,
+                              size,
+                              name,
+                              flags,
+                              callback,
+                              user_data);
+}
+
+/**
+ * gdu_drive_create_volume_finish:
+ * @drive: A #GduDrive.
+ * @res: A #GAsyncResult.
+ * @error: A #GError or %NULL.
+ *
+ * Finishes an operation started with gdu_drive_create_volume().
+ *
+ * Returns: A #GduVolume for the created volume or %NULL if @error is
+ * set. The returned object must be freed with g_object_unref().
+ */
+GduVolume *
+gdu_drive_create_volume_finish (GduDrive              *drive,
+                                GAsyncResult          *res,
+                                GError               **error)
+{
+        GduDriveClass *klass = GDU_DRIVE_GET_CLASS (drive);
+        return klass->create_volume_finish (drive,
+                                            res,
+                                            error);
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static void
+get_volumes_add_enclosed (GduPresentable  *p,
+                          GList          **ret)
+{
+        GList *enclosed;
+        GList *l;
+        GduPool *pool;
+
+        pool = gdu_presentable_get_pool (p);
+        enclosed = gdu_pool_get_enclosed_presentables (pool, p);
+
+        for (l = enclosed; l != NULL; l = l->next) {
+                GduPresentable *ep = GDU_PRESENTABLE (l->data);
+
+                if (!GDU_IS_VOLUME (ep))
+                        continue;
+
+                get_volumes_add_enclosed (ep, ret);
+
+                *ret = g_list_prepend (*ret, g_object_ref (p));
+        }
+
+        g_list_foreach (enclosed, (GFunc) g_object_unref, NULL);
+        g_list_free (enclosed);
+        g_object_unref (pool);
+}
+
+/**
+ * gdu_drive_get_volumes:
+ * @drive: A #GduDrive
+ *
+ * Returns a list of all #GduVolume<!-- -->s for @drive.
+ *
+ * Returns: A #GList of #GduVolume objects. The caller must free the list and each element.
+ */
+GList *
+gdu_drive_get_volumes (GduDrive *drive)
+{
+        GList *ret;
+
+        /* TODO: do we need a vfunc for this? */
+
+        ret = NULL;
+
+        get_volumes_add_enclosed (GDU_PRESENTABLE (drive), &ret);
+
+        ret = g_list_reverse (ret);
+
+        return ret;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
